import requests
import gzip
import tqdm
import os
import importlib.resources
import pickle
import struct
from tqdm import tqdm

def download_cache(cache_url : str, target_file : str) -> None:
    '''
    This function takes the URL of a cache file, downloads it (with a progress bar) and saves
    it to the target_file. If the URL extension ends with .gz, the file is unzipped as it
    downloads
    '''

    # If the URL is incomplete, prepend https://
    if not (cache_url.startswith('http://') or cache_url.startswith('https://')):
        cache_url = 'https://' + cache_url

    # Check whether the file is a .gz file
    is_gz = cache_url.endswith('.gz')

    # Start the request
    with requests.get(cache_url, stream=True) as response:
        # Raise an error if the request was unsuccessful
        response.raise_for_status()

        # Get the total size and prepare the progress bar
        total_size = int(response.headers.get('content-length', 0))
        progress_bar = tqdm.tqdm(total=total_size, unit='B', unit_scale=True, desc='Downloading cache file')

        if is_gz:
            # Monkey patch the read method
            original_read = response.raw.read
            def new_read(chunk_size=-1):
                data = original_read(chunk_size)
                if data:
                    progress_bar.update(len(data))
                return data
            response.raw.read = new_read

            with gzip.GzipFile(fileobj=response.raw, mode='rb') as gz_in, open(target_file, 'wb') as f:
                while True:
                    data = gz_in.read(8192)
                    if not data:
                        break
                    f.write(data)

        else:
            with open(target_file, 'wb') as f:
                for chunk in response.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)
                        progress_bar.update(len(chunk))
        
        progress_bar.close()

class TqdmFileReader:
    '''
    This file reader class wraps a file object and provides a tqdm progress bar
    '''

    def __init__(self, file_path):
        self.f = open(file_path, "rb")
        self.total = os.path.getsize(file_path)
        self.pbar = tqdm(total=self.total, unit='B', unit_scale=True)
    
    def read(self, size=-1):
        chunk = self.f.read(size)
        self.pbar.update(len(chunk))
        return chunk

    def readline(self, *args, **kwargs):
        line = self.f.readline(*args, **kwargs)
        self.pbar.update(len(line))
        return line

    def __getattr__(self, attr):
        return getattr(self.f, attr)

    def close(self):
        self.f.close()
        self.pbar.close()

def load_cache_file(cache_file : bytes) -> dict:
    '''
    This function accept a bytes string containing a pickled object; if first checks
    whether it can be loaded uncompressed, and if not, it tries to load it as a
    compressed file
    
    It returns the object from the pickle
    '''

    try:
        print('Loading, please wait...')
        reader = TqdmFileReader(cache_file)
        cache = pickle.load(reader)
        reader.close()

        return cache
    except:
        try:
            return pickle.loads(gzip.decompress(cache_file))
        except:
            raise ValueError('Invalid cache file provided')

def get_cache(cache_file_name      : str          ,
              temp_cache_file_name : str          ,
              dev_mode             : bool         ,
              delay_responses_new  : bool         ,
              verbose              : bool = False  ) -> tuple[bool, dict]:
    '''
    This function attempts to load the cache from disk, in the following order of priority
      - First, we look in the working directory
      - Then, we look in the directory containing the package
      - Finally, two options
          * If we are in dev mode, we create a blank cache
          * Otherwise, we prompt the user for a URL to download the cache therefrom 
            (compressed or not). It saves it in the working directory with the name
            cache_file_name
    
    If we are in dev_mode, additionally carry out the following tasks
      - Try and load the temporary cache file, if it exists, and update the main cache
        with it
      - If the delay_responses parameter has changed (i.e., if delay_responses in the cache
        file is not equal to delay_responses_new), it is overwritten in the cache file
    
    It returns a tuple with two entries
      - The value of delay_responses to use
      - The cache
    '''

    # Load the existing cache
    # -----------------------

    # Get the location of the package in case we need it later
    package_location = importlib.resources.files(__package__)

    # Determine the location of the cache file
    if os.path.exists(cache_file_name):
        # First, look in the working directory
        cache_loc = cache_file_name
        
        if verbose:
            print('Cache file found in the working directory.')

    elif os.path.exists(package_location.joinpath(cache_file_name)):
        # If there is no cache file in the working directory, look in the package
        # directory
        cache_loc = package_location.joinpath(cache_file_name)
        
        if verbose:
            print('Cache file found in the package directory.')

    else:
        # If we have neither and we are in dev mode, start with an empty cache; if not
        # ask the user whether they want to download a cache file, and download it to the
        # working directory
        if dev_mode:
            print('No cache file was found; starting with an empty cache. It will be saved in '
                    f'your working directory, with the name {cache_file_name}.')
            cache_loc = cache_file_name
            pickle.dump((delay_responses_new, {}), open(cache_loc, 'wb'))

        else:
            raise(BaseException('\nERROR: No cache file was found in the folder in which this notebook is located.\n'
                                  "Please make sure you've downloaded the files from the start of this module and\n"
                                  'that you are running this notebook in the downloaded folder. Please contact\n'
                                  'the teaching team if you are still having issues.\n'))
            
            cache_url = input('No cache file found; please enter a URL to download a cache file from\n')

            try:
                download_cache(cache_url, cache_file_name)
                cache_loc = cache_file_name
                
                print('Cache file downloaded successfully.')
            except:
                print('Failed to download cache file')
                raise
    
    # Finally, read the cache file
    with open(cache_loc, 'rb') as f : data = f.read()
    delay_responses, cache = load_cache_file(data)
    
    # Update with the temporary cache file, or the new delay_responses value
    # ----------------------------------------------------------------------

    # If we are in dev mode, load any temporary cache file, and update the cache with it; then,
    # re-save the updated cache
    if dev_mode:
        if os.path.exists(temp_cache_file_name):
            if verbose:
                print('Temporary cache file found; using it to update the cache.')
            
            # Load the temporary cache file
            with open(temp_cache_file_name, 'rb') as f:
                while True:
                    # Read the length of the next entry; break if we're reached the end of
                    # the file
                    length_data = f.read(4)
                    if not length_data:
                        break
                    length = struct.unpack('I', length_data)[0]

                    key, value = pickle.loads(f.read(length))

                    cache[key] = value

            # Re-save the cache (note that in dev mode, we always save it uncompressed)
            # Use the new delay_responses value
            pickle.dump((delay_responses_new, cache), open(cache_loc, 'wb'))

            # Delete the temporary cache file
            os.remove(temp_cache_file_name)

            if verbose:
                print('Cache updated and re-saved. Temporary cache file deleted')

        elif delay_responses_new != delay_responses:
            pickle.dump((delay_responses_new, cache), open(cache_loc, 'wb'))

            if verbose:
                print('delay_responses value was changed - cache re-saved')

        # Set the value of delay_responses to the new value
        delay_responses = delay_responses_new

    # Return
    # ------
    return delay_responses, cache


################################
#   Materialization utilities  #
################################

import gzip
import pickle
import base64
import hashlib
import importlib

# This file contains utilities to materialize the cache into a form that
# can be distributed

def materialize_cache(cache           : dict,
                      delay_responses : bool,
                      compress        : bool,
                      hash_keys       : bool,
                      file_name       : str | None,
                      used_keys       : list | None = None) -> str | None:
    '''
    This function materializes the cache into a pickle file for distribution. It accepts
    the following arguments:
      - cache : the cache to be saved
      - compress : whether the cache should be compressed
      - hash_keys : whether keys should be hashed; if yes, the dictionary keys will be hashed
                    and the pickle file will be smaller. On the other hand, the cache will then
                    become *final* - it will be impossible to add to it. If you choose to go
                    the hashed direction, it is recommended you first save the cache without
                    hashing the keys so that you can later add to it if you like
      - file_name : the file name to be used. If no filename is provided, a base64 encoded
                    string is returned
      - used_keys : a list of keys that should be included in the cache; any keys are that are
                    not included in this list are discarded
    '''

    # If we want used keys only, filter down the cache
    if used_keys:
        cache = {i:j for i, j in cache.items() if i in used_keys}

    # If we want to hash the keys, go ahead and do it
    if hash_keys:
        cache = {hashlib.md5(i.encode('utf-8')).hexdigest():j
                                        for i, j in cache.items()}
        
        # Make all pointers point to hashed values
        for key, vals in cache.items():
            for val in vals:
                if 'TARGET' in val:
                    val['TARGET'] = hashlib.md5(val['TARGET'].encode('utf-8')).hexdigest()

    # Pickle the dictionary
    pickled_data = pickle.dumps((delay_responses, cache))

    # Compress it if needed
    if compress:
        pickled_data = gzip.compress(pickled_data)
    
    # If the filename is provided, save it there; otherwise, return a b64 encoded
    # string
    if file_name:
        with open(file_name, 'wb') as f:
            f.write(pickled_data)
    
    else:
        return base64.b64encode(pickled_data).decode('utf-8')
    
def create_self_contained(cache           : dict,
                          delay_responses : bool,
                          compress        : bool,
                          hash_keys       : bool,
                          file_name       : str | None,
                          used_keys       : list | None = None) -> str | None:
    '''
    Returns a self-contained .py file that includes the cache inside of it.

    See _materialized_cache for an explanation of the arguments
    '''

    # Create a placeholder for the output code
    out_code = []

    # Get the cache as a base 64 encoded string; do not include the file name to
    # ensure nothing is saved
    b64_cache = materialize_cache(cache, delay_responses, compress, hash_keys, None, used_keys)

    # Ensure only the OpenAI and AsyncOpenAI functions are exposed
    out_code.append("__all__ = ['OpenAI', 'AsyncOpenAI']")

    # Add the cache and the code to decompress it to the self-contained file
    out_code.append('import base64')
    out_code.append('import pickle')
    out_code.append(f'cache = base64.b64decode("{b64_cache}")')

    if compress:
        out_code.append('import gzip')
        out_code.append('cache = gzip.decompress(cache)')
        
    out_code.append('DELAY_RESPONSES, cache = pickle.loads(cache)')

    # Load the cached_client.py file
    with open(importlib.resources.files(__package__).joinpath('cached_client.py'), 'r') as f:
        out_code.extend(f.read().split('\n'))
    
    # Create the main entrypoints
    entrypoint_args = ( '                  api_key          ,'
                        'cache           = cache            ,'
                        'verbose         = False            ,'
                        'dev_mode        = False            ,'
                        'delay_responses = DELAY_RESPONSES  ,'
                        'temp_cache_file = None             ,'
                        'used_keys_file  = None             ')
    
    out_code.append(f'def OpenAI     (api_key : str | None = None) : return CachedClient({entrypoint_args}, is_async = False)')
    out_code.append(f'def AsyncOpenAI(api_key : str | None = None) : return CachedClient({entrypoint_args}, is_async = True)' )

    # Save the result
    with open(file_name, 'w') as f:
        f.write('\n'.join(out_code))


# File utilities
import os
from . import utils
from . import cached_client
from . import materialize_utils
import datetime

######################
#  Define constants  #
######################

# Specify the name of the permanent cache file. We will first look for this file
# in the working directory; if it exists there, we load it from there. If it
# doesn't, we look for it in the package directory
CACHE_FILE_NAME       = 'openai.cache'

# Determine whether we want to print debugging information as we run the package;
# this is required if the 'CACHED_OPENAI_VERBOSE' variable is set in the
# environment
VERBOSE               = 'CACHED_OPENAI_VERBOSE' in os.environ

# In dev mode, all calls to the open AI API will be cached in the temporary cache
# file, and every time a request is made (either new or cached), we will track
# the fact the request has been made in a used keys file. The latter allows a
# "cleaning operation" that only keeps data for requests that have been made
# since this file was created.
#
# Dev mode is turned on if the 'CACHED_OPENAI_DEV_MODE' variable is set in the
# environment
DEV_MODE              = 'CACHED_OPENAI_DEV_MODE' in os.environ
if VERBOSE:
    print(f'DEV_MODE is set to {DEV_MODE}')

# The name of the temporary cache and used keys files which will be written to in
# dev mode. These will be in the code's working directory
TEMP_CACHE_FILE_NAME  = 'openai_cache_temp.bin' 
USED_KEYS_FILE        = 'openai_cache_used.txt'

# When we recover values form the cache, we can either return the values instantly
# or delay for as long as it took to originally get a response from the OpenAI
# API. By default, no pausing will happen, unless a 'CACHED_OPENAI_DELAY_RESPONSES'
# variable is set in the environment 
DELAY_RESPONSES_NEW   = 'CACHED_OPENAI_DELAY_RESPONSES' in os.environ
if VERBOSE:
    print(f'DELAY_RESPONSES is set to {DELAY_RESPONSES_NEW}')

###########
#  Setup  #
###########

# Load the cache
delay_responses, cache = utils.get_cache(CACHE_FILE_NAME, TEMP_CACHE_FILE_NAME, DEV_MODE, DELAY_RESPONSES_NEW, VERBOSE)

#################################################
#  Create the main entrypoints for the package  #
#################################################

def OpenAI(api_key : str | None = None):
    return cached_client.CachedClient(                  api_key                   ,
                                      cache           = cache                     ,
                                      verbose         = VERBOSE                   ,
                                      dev_mode        = DEV_MODE                  ,
                                      is_async        = False                     ,
                                      delay_responses = delay_responses           ,
                                      temp_cache_file = TEMP_CACHE_FILE_NAME      ,
                                      used_keys_file  = USED_KEYS_FILE              )

def AsyncOpenAI(api_key : str | None = None):
    return cached_client.CachedClient(                 api_key                    ,
                                      cache           = cache                     ,
                                      verbose         = VERBOSE                   ,
                                      dev_mode        = DEV_MODE                  ,
                                      is_async        = True                      ,
                                      delay_responses = delay_responses           ,
                                      temp_cache_file = TEMP_CACHE_FILE_NAME      ,
                                      used_keys_file  = USED_KEYS_FILE              )

##########################################
#  Create the materialization functions  #
##########################################

def materialize(self_contained : bool,
                compress       : bool,
                hash_keys      : bool,
                used_keys_only : bool):
    '''
    This function materilizes the cache and prepares it for distribution. It accepts
    the following options
      - self_contained : if True, the function will create a file called
        cached_openai_{date}.py in the current working directory which will contain
        the materialized cache. This file will be a self-contained .py file containing
        a serialized verison of the cache. Just import that package, and you're good
        to go with the full cache.

        If False, the function will create a file called named openai_{date}.cache
        containing the cache only. The file can be distributed in three ways
          - Rename it to openai.cache, and ask your target audience to place the file
            in their working directory. They can then install cached_openai through
            pypi, and when they load the package, it will find the file in the working
            directory and load it
          - Upload it to some URL. Then, get your audience to install cached_openai
            from pypi. When they load the package, it will prompt them for the URL,
            from which the file will autoomatically be downloaded
          - Fork cached_openai, rename the file to openai.cache, and include it in
            the src directory of the package. Publish it. When your target audience
            downloads the package, the cache will be read from this file.

      - compress: True if the cache should be compressed before saving or false
        otherwise. Note that this means the package will take longer to run every time
        it is loaded 
      
      - hash_keys : whether keys should be hashed; if True, the dictionary keys will be hashed
        and the pickle file will be smaller. On the other hand, the cache will then become
        *final* - it will be impossible to add to it. If you choose to go the hashed direction,
        it is recommended you first save the cache without hashing the keys so that you can
        later add to it if you like

      - used_keys_only : if True, the cache will only contain the keys that have been
        created or access since the last time the openai_cache_used.txt file was created.

        This is useful if you fill the cache with a whole bunch of data while you test
        your code, and then want to create a "clean" version to distribute. Just delete the
        openai_cache_used.txt file, run you clean code, and then run this with used_keys_only
        = True.
    '''

    current_date = datetime.datetime.now().strftime('%Y-%m-%d')

    if used_keys_only:
        with open(USED_KEYS_FILE, 'r') as f:
            used_keys = f.read().split('\n')
    else:
        used_keys = None

    if self_contained:
        materialize_utils.create_self_contained(cache           = cache,
                                                delay_responses = delay_responses,
                                                compress        = compress,
                                                hash_keys       = hash_keys,
                                                file_name       = f'cached_openai_{current_date}.py',
                                                used_keys       = used_keys)
    else:
        materialize_utils.materialize_cache(cache           = cache,
                                            delay_responses = delay_responses,
                                            compress        = compress,
                                            hash_keys       = hash_keys,
                                            file_name       = CACHE_FILE_NAME.split('.')[0]
                                                                    + f'_{current_date}.'
                                                                        + CACHE_FILE_NAME.split('.')[-1],
                                            used_keys       = used_keys)


# Make the main entrypoints into the package available at the top level of the package
from .main import OpenAI, AsyncOpenAI, DEV_MODE

# Make the materialization functions available at the top level of the package, if
# we're in dev mode
if DEV_MODE:
    from .main import materialize

# Remove DEV_MODE from the namespace
del DEV_MODE